# Copyright (C) 2026 pytest-qfield Contributors.
#
#
# This file is part of pytest-qfield.
#
# pytest-qfield is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# pytest-qfield is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pytest-qfield.  If not, see <https://www.gnu.org/licenses/>.

from typing import TYPE_CHECKING, Any, cast

from PyQt6.QtCore import (
    QObject,
    QSettings,
    QVariant,
    pyqtProperty,
    pyqtSignal,
    pyqtSlot,
)
from qgis.core import QgsRasterLayer, QgsVectorLayer

if TYPE_CHECKING:
    from qgis.core import QgsFeature, QgsGeometry, QgsMapLayer, QgsProject


class QgsMapLayerStub(QObject):
    """
    Stub implementation of QgsMapLayer.
    """

    def __init__(self, qgis_layer: "QgsMapLayer") -> None:
        super().__init__(parent=qgis_layer)
        self.qgis_layer = qgis_layer

    @pyqtProperty(str)
    def name(self) -> str:
        return self.qgis_layer.name()

    @pyqtProperty(bool)
    def isValid(self) -> bool:
        # For some reason isValid is not a function,
        # but an attribute (or a property in python)
        return self.qgis_layer.isValid()

    @staticmethod
    def create_from_qgs_map_layer(map_layer: "QgsMapLayer") -> "QgsMapLayerStub":
        if isinstance(map_layer, QgsVectorLayer):
            return QgsVectorLayerStub(map_layer)
        if isinstance(map_layer, QgsRasterLayer):
            return QgsRasterLayerStub(map_layer)
        raise NotImplementedError


class QgsVectorLayerStub(QgsMapLayerStub):
    """
    Stub implementation of QgsVectorLayer.
    """

    featureAdded = pyqtSignal(int)

    def __init__(self, qgis_layer: "QgsVectorLayer") -> None:
        super().__init__(qgis_layer)
        self.qgis_layer: QgsVectorLayer = cast("QgsVectorLayer", self.qgis_layer)
        self.qgis_layer.featureAdded.connect(self.featureAdded.emit)

    @pyqtSlot(result=bool)
    def startEditing(self) -> bool:
        return self.qgis_layer.startEditing()

    @pyqtSlot(result=bool)
    def commitChanges(self) -> bool:
        # TODO: stop editing or not?
        return self.qgis_layer.commitChanges(stopEditing=False)

    @pyqtSlot(int, result=QObject)
    def getFeature(self, feature_id: int) -> "QgsFeatureStub":
        stub = QgsFeatureStub(self.qgis_layer.getFeature(feature_id))
        stub.setParent(self)
        return stub


class QgsRasterLayerStub(QgsMapLayerStub):
    """
    Stub implementation of QgsRasterLayer.
    """

    featureAdded = pyqtSignal(int)

    def __init__(self, qgis_layer: "QgsRasterLayer") -> None:
        super().__init__(qgis_layer)
        self.qgis_layer: QgsRasterLayer = cast("QgsRasterLayer", self.qgis_layer)


class QgsProjectStub(QObject):
    """
    Stub implementation for QgsProject (qgisProject in QML).
    """

    def __init__(self, qgis_project: "QgsProject") -> None:
        super().__init__(parent=qgis_project)
        self.qgis_project = qgis_project

    @pyqtSlot(str, result=list)
    def mapLayersByName(self, name: str) -> list[QgsVectorLayerStub]:
        return list(
            map(
                QgsMapLayerStub.create_from_qgs_map_layer,
                self.qgis_project.mapLayersByName(name),
            )
        )


class QSettingsStub(QSettings):
    @pyqtSlot(str, result="QVariant")
    @pyqtSlot(str, "QVariant", result="QVariant")
    def value(self, key: str, default_value: Any = None) -> "QVariant":  # noqa: ANN401
        return super().value(key, default_value)


class QgsFeatureStub(QObject):
    def __init__(self, qgis_feature: "QgsFeature") -> None:
        super().__init__(parent=None)
        self.qgis_feature = qgis_feature

    @pyqtProperty(int)
    def id(self) -> int:
        return self.qgis_feature.id()

    @pyqtSlot(str, result="QVariant")
    def attribute(self, attribute_name: str) -> str | int | float | bool | None:
        return self.qgis_feature[attribute_name]

    @pyqtSlot(str, QVariant)
    def setAttribute(self, attribute_name: str, raw_value: QVariant | str) -> None:
        value = raw_value.value() if isinstance(raw_value, QVariant) else raw_value
        self.qgis_feature.setAttribute(
            self.qgis_feature.fieldNameIndex(attribute_name), value
        )

    @pyqtProperty(QObject)
    def geometry(self) -> "QgsGeometryStub":
        geometry_stub = QgsGeometryStub(self.qgis_feature.geometry())
        geometry_stub.setParent(self)
        return geometry_stub


class QgsGeometryStub(QObject):
    def __init__(self, qgis_geometry: "QgsGeometry") -> None:
        super().__init__(parent=None)
        self.qgis_geometry = qgis_geometry

    @pyqtSlot(int, result=str)
    def asWkt(self, precision: int) -> str:
        return self.qgis_geometry.asWkt(precision)



# Copyright (C) 2026 pytest-qfield Contributors.
#
#
# This file is part of pytest-qfield.
#
# pytest-qfield is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# pytest-qfield is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pytest-qfield.  If not, see <https://www.gnu.org/licenses/>.
import uuid
from typing import TYPE_CHECKING

from PyQt6.QtCore import QObject, QSizeF, pyqtSignal, pyqtSlot
from PyQt6.QtQuick import QQuickItem
from qgis.core import QgsFeatureRequest, QgsGeometry, QgsVectorLayerUtils

from pytest_qfield.stub_interface.qgis_stubs import (
    QgsFeatureStub,
    QgsGeometryStub,
    QgsVectorLayerStub,
)

if TYPE_CHECKING:
    from qgis.gui import QgisInterface


class QFieldAppInterfaceStub(QObject):
    """
    Stub implementation of AppInterface.

    https://api.qfield.org/QField/classAppInterface/
    """

    loadProjectEnded = pyqtSignal()

    def __init__(self, qgis_iface: "QgisInterface") -> None:
        super().__init__(parent=None)
        self.added_item_count = 0
        self.logged_messages: list[str] = []
        self.qgis_iface = qgis_iface

        self._qgis_main_window: QObject | None = None
        self._qml_main_window: QObject | None = None
        self.qgis_map_canvas = qgis_iface.mapCanvas()

    @property
    def toast_messages(self) -> list[str]:
        toast_messages = self.qml_main_window.property("toastMessages")
        if hasattr(toast_messages, "toVariant"):
            return toast_messages.toVariant()
        return toast_messages

    @property
    def qml_main_window(self) -> QObject:
        if self._qml_main_window is None:
            raise ValueError("Add QML main window via set_main_window method first!")
        return self._qml_main_window

    def set_main_window(
        self, qgis_main_window: QObject, qml_main_window: QObject
    ) -> None:
        self._qgis_main_window = qgis_main_window
        self._qml_main_window = qml_main_window

    def show(self) -> None:
        if self._qgis_main_window is None:
            raise ValueError("Add mainWindow via set_main_window method first!")
        self._qgis_main_window.show()

    def set_qml_root(self, root: QObject) -> None:
        if not isinstance(root, QQuickItem):
            raise TypeError(f"Unsupported root type: {type(root)}")

        scene_root = self.qml_main_window.findChild(QObject, "host")
        if scene_root is None:
            raise RuntimeError("Could not find host item in main window")
        if not isinstance(scene_root, QQuickItem):
            raise TypeError(f"Host is not a QQuickItem: {type(scene_root)}")

        root.setParentItem(scene_root)
        root.setSize(scene_root.size())

    @pyqtSlot(result=QObject)
    def mainWindow(self) -> QObject:
        return self.qml_main_window

    @pyqtSlot(result=QObject)
    def mapCanvas(self) -> QObject:
        return self.qml_main_window

    @pyqtSlot(QObject)
    def addItemToPluginsToolbar(self, _item: "QQuickItem") -> None:
        toolbar_row = self.qml_main_window.findChild(QObject, "pluginsToolbarRow")
        if toolbar_row is None:
            raise RuntimeError("Plugins toolbar row not found")
        _item.setSize(QSizeF(48, 48))
        _item.setParentItem(toolbar_row)

    @pyqtSlot(str)
    @pyqtSlot(str, str)
    @pyqtSlot(str, str, str)
    @pyqtSlot(str, str, str, str)
    @pyqtSlot(str, str, str, str, str)
    @pyqtSlot(str, str, str, str, str, str)
    @pyqtSlot(str, str, str, str, str, str, str)
    @pyqtSlot(str, str, str, str, str, str, str, str)
    @pyqtSlot(str, str, str, str, str, str, str, str, str)
    @pyqtSlot(str, str, str, str, str, str, str, str, str, str)
    def logMessage(self, *messages: str) -> None:
        self.logged_messages.extend(messages)


class QFieldPlatformUtilitiesStub(QObject):
    @pyqtSlot(result=bool)
    def isSystemDarkTheme(self) -> bool:
        return False


class QFieldStringUtilsStub(QObject):
    """
    StringUtils stub.

    https://api.qfield.org/QField/classStringUtils/
    """

    @pyqtSlot(result=str)
    def createUuid(self) -> str:
        return f"{{{uuid.uuid4()}}}"


class QFieldGeometryUtilsStub(QObject):
    """
    GeometryUtils stub

    https://api.qfield.org/QField/classGeometryUtils/
    """

    # TODO: add point method

    @pyqtSlot(str, result=QObject)
    def createGeometryFromWkt(self, wkt: str) -> QgsGeometryStub:
        geometry_stub = QgsGeometryStub(QgsGeometry.fromWkt(wkt))
        geometry_stub.setParent(self)
        return geometry_stub


class QFieldLayerUtilsStub(QObject):
    """
    LayerUtils stub.

    https://api.qfield.org/QField/classLayerUtils
    """

    @pyqtSlot(QObject, QObject, result=bool)
    def addFeature(self, layer: QgsVectorLayerStub, feature: QgsFeatureStub) -> bool:
        return layer.qgis_layer.addFeature(feature.qgis_feature)

    @pyqtSlot(QObject, str, result=QObject)
    def createFeatureIteratorFromExpression(
        self,
        layer: QgsVectorLayerStub,
        expression: str,
    ) -> "FeatureIteratorStub":
        iterator_stub = FeatureIteratorStub(
            layer, QgsFeatureRequest().setFilterExpression(expression)
        )
        iterator_stub.setParent(self)
        return iterator_stub

    def get_iterators(self) -> list["FeatureIteratorStub"]:
        return self.findChildren(FeatureIteratorStub)

    def clear_iterators(self) -> None:
        for child in self.findChildren(FeatureIteratorStub):
            child.setParent(None)
            child.deleteLater()


class FeatureIteratorStub(QObject):
    """FeatureIterator stub."""

    def __init__(self, layer: QgsVectorLayerStub, request: QgsFeatureRequest) -> None:
        super().__init__(parent=None)
        self.layer = layer
        self.request = request
        self._features = list(layer.qgis_layer.getFeatures(self.request))
        self.iterated_count = 0
        self.closed = False

    @pyqtSlot(result=bool)
    def hasNext(self) -> bool:
        return len(self._features) > 0 and self.iterated_count < len(self._features)

    @pyqtSlot(result=QObject)
    def next(self) -> QgsFeatureStub:
        feature_stub = QgsFeatureStub(self._features[self.iterated_count])
        feature_stub.setParent(self)
        self.iterated_count += 1
        return feature_stub

    @pyqtSlot()
    def close(self) -> None:
        self.closed = True


class QFieldFeatureUtilsStub(QObject):
    """FeatureUtils stub.

    https://api.qfield.org/QField/classFeatureUtils/
    """

    @pyqtSlot(QObject, result=QObject)
    @pyqtSlot(QObject, QObject, result=QObject)
    def createFeature(
        self, layer: QgsVectorLayerStub, geometry: QgsGeometryStub | None = None
    ) -> QgsFeatureStub:
        feature = QgsVectorLayerUtils.createFeature(
            layer.qgis_layer,
            geometry=geometry.qgis_geometry if geometry else QgsGeometry(),
        )
        stub = QgsFeatureStub(feature)
        stub.setParent(self)
        return stub



# Copyright (C) 2026 pytest-qfield Contributors.
#
#
# This file is part of pytest-qfield.
#
# pytest-qfield is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# pytest-qfield is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pytest-qfield.  If not, see <https://www.gnu.org/licenses/>.
import os
from collections.abc import Iterator
from pathlib import Path
from typing import TYPE_CHECKING

import pytest
from PyQt6.QtCore import QObject, Qt, QUrl
from PyQt6.QtQml import (
    QQmlApplicationEngine,
    qmlRegisterType,
)
from PyQt6.QtQuickWidgets import QQuickWidget
from PyQt6.QtWidgets import QSizePolicy, QVBoxLayout, QWidget
from qgis.core import QgsProject

from pytest_qfield.qfieldbot import QFieldBot
from pytest_qfield.stub_interface.qfield_stubs import (
    QFieldAppInterfaceStub,
    QFieldFeatureUtilsStub,
    QFieldGeometryUtilsStub,
    QFieldLayerUtilsStub,
    QFieldPlatformUtilitiesStub,
    QFieldStringUtilsStub,
)
from pytest_qfield.stub_interface.qgis_stubs import QgsProjectStub, QSettingsStub

if TYPE_CHECKING:
    from _pytest.config import Parser
    from _pytest.fixtures import SubRequest
    from PyQt6.QtWidgets import QApplication
    from pytestqt.logging import _QtMessageCapture
    from pytestqt.qtbot import QtBot
    from qgis.gui import QgisInterface, QgsMapCanvas

_QFIELD_IMPORTS_DIR_ENV = "QFIELD_IMPORTS_DIR"
_QFIELD_IMPORT_DIR_KEY = "qfield_imports_dir"


@pytest.fixture(scope="session")
def qapp_args() -> list[str]:
    return ["QField test"]


@pytest.fixture
def qfield_new_project(qgis_new_project: "QgsProject") -> None:  # noqa: ARG001
    """Initialize new QField project"""
    return


@pytest.fixture
def qfield_bot(  # noqa: PLR0913
    qgis_app: "QApplication",
    qgis_parent: QWidget,
    qgis_canvas: "QgsMapCanvas",
    qfield_iface: QFieldAppInterfaceStub,
    qfield_platform_utilities_stub: QFieldPlatformUtilitiesStub,
    qgs_project_stub: QgsProjectStub,
    qfield_string_utils_stub: QFieldStringUtilsStub,
    qfield_layer_utils_stub: QFieldLayerUtilsStub,
    qfield_feature_utils_stub: QFieldFeatureUtilsStub,
    qfield_geometry_utils_stub: QFieldGeometryUtilsStub,
    register_qfield_resources: None,  # noqa: ARG001
    register_qfield_types: None,  # noqa: ARG001
    register_qgis_types: None,  # noqa: ARG001
    main_window_qml_path: Path,
    qtbot: "QtBot",
    qtlog: "_QtMessageCapture",
    request: "SubRequest",
    qfield_qml_extra_context_properties: dict[str, object],
    tmp_path: "Path",
) -> Iterator["QFieldBot"]:
    """Fixture used to create a QFieldBot instance for use during testing."""
    qfield_import_path = _get_qfied_import_path(request)

    engine = QQmlApplicationEngine()

    # Load QField imports
    engine.addImportPath(str(qfield_import_path))

    # Inject QField interface stubs
    qfield_iface.setParent(engine)
    qfield_platform_utilities_stub.setParent(engine)

    system_font_point_size = qgis_app.font().pointSizeF() + 2.0

    # Inject context properties
    context_properties = {
        "iface": qfield_iface,
        "platformUtilities": qfield_platform_utilities_stub,
        "qgisProject": qgs_project_stub,
        "systemFontPointSize": system_font_point_size,
        "settings": QSettingsStub(),
        "StringUtils": qfield_string_utils_stub,
        "LayerUtils": qfield_layer_utils_stub,
        "FeatureUtils": qfield_feature_utils_stub,
        "GeometryUtils": qfield_geometry_utils_stub,
        **qfield_qml_extra_context_properties,
    }

    for property_name, property_value in context_properties.items():
        engine.rootContext().setContextProperty(property_name, property_value)

    qfield_bot = QFieldBot(
        qml_engine=engine,
        qfield_iface=qfield_iface,
        qtbot=qtbot,
        qtlog=qtlog,
        tmp_path=tmp_path,
    )

    # Load and embed the QML shell in the pytest-qgis main window.
    qml_overlay_widget, qml_main_window = _load_qml_overlay_widget(
        qml_engine=engine,
        main_window_qml_path=main_window_qml_path,
        parent=qgis_parent,
    )

    _embed_qml_window_in_qgis_main_window(
        qgis_parent=qgis_parent,
        qgis_canvas=qgis_canvas,
        qml_overlay_widget=qml_overlay_widget,
    )

    qfield_iface.set_main_window(
        qgis_main_window=qgis_parent,
        qml_main_window=qml_main_window,
    )
    # Yielding to keep stub instances alive
    yield qfield_bot  # noqa: PT022


@pytest.fixture
def main_window_qml_path() -> Path:
    """
    Path to a file with QML Code of the MainWindow

    Feel free to override this fixture to use a custom QML code for the MainWindow.
    """
    main_window_qml_path = Path(__file__).parent / "main_window.qml"
    if not main_window_qml_path.exists():
        raise FileNotFoundError(
            f"QML path for main_window.qml cannot be found in {main_window_qml_path}!"
        )

    return main_window_qml_path


@pytest.fixture
def qfield_iface(qgis_iface: "QgisInterface") -> QFieldAppInterfaceStub:
    """
    Stub implementation for QFieldAppInterface.

    Override this fixture to use an extended version of the class if needed.
    """
    return QFieldAppInterfaceStub(qgis_iface)


@pytest.fixture
def qfield_platform_utilities_stub() -> QFieldPlatformUtilitiesStub:
    """
    Stub implementation for QFieldPlatformUtilities.

    Override this fixture to use an extended version of the class if needed.
    """
    return QFieldPlatformUtilitiesStub()


@pytest.fixture
def qgs_project_stub() -> QgsProjectStub:
    """
    Stub implementation for QgsProject (qgisProject in QML).

    Override this fixture to use an extended version of the class if needed.
    """
    return QgsProjectStub(QgsProject.instance())


@pytest.fixture
def qfield_string_utils_stub() -> QFieldStringUtilsStub:
    """
    Stub implementation for StringUtils.

    Override this fixture to use an extended version of the class if needed.
    """
    return QFieldStringUtilsStub()


@pytest.fixture(autouse=True)
def qfield_layer_utils_stub() -> Iterator[QFieldLayerUtilsStub]:
    """
    Stub implementation for LayerUtils.

    Override this fixture to use an extended version of the class if needed.
    """
    layer_utils_stub = QFieldLayerUtilsStub()
    yield layer_utils_stub
    iterators = layer_utils_stub.get_iterators()
    try:
        for iterator in iterators:
            if not iterator.closed:
                raise AssertionError("Iterator was not closed")
    finally:
        layer_utils_stub.clear_iterators()


@pytest.fixture
def qfield_feature_utils_stub() -> QFieldFeatureUtilsStub:
    """
    Stub implementation for FeatureUtils.

    Override this fixture to use an extended version of the class if needed.
    """
    return QFieldFeatureUtilsStub()


@pytest.fixture
def qfield_geometry_utils_stub() -> QFieldGeometryUtilsStub:
    """
    Stub implementation for GeometryUtils.

    Override this fixture to use an extended version of the class if needed.
    """
    return QFieldGeometryUtilsStub()


@pytest.fixture
def qfield_qml_extra_context_properties() -> dict[str, object]:
    """
    Override this fixture to provide extra context properties for QField QML files.
    """
    return {}


@pytest.fixture
def register_qfield_resources() -> None:
    """
    Registers compiled resources.
    Override this fixture to provide your own compiled resources.
    """
    # Register /gml
    import pytest_qfield.qfield_resources  # noqa: F401, PLC0415


@pytest.fixture(scope="session")
def register_qfield_types() -> None:
    """
    Registers qfield types used in QField plugin.
    Override this fixture to provide more types needed.

    Check this file for a complete list
    (lines that register something to "org.qfield")
    https://github.com/opengisch/QField/blob/master/src/core/qgismobileapp.cpp
    """
    # Something provided here just to get rid of warnings
    qmlRegisterType(QgsProjectStub, "org.qfield", 1, 0, "Foo")
    # qmlRegisterUncreatableType
    # qmlRegisterSingletonType


@pytest.fixture(scope="session")
def register_qgis_types() -> None:
    """
    Registers qgus types used in QField plugin.
    Override this fixture to provide more types needed.

    Check this file for a complete list
    (lines that register something to "org.qgis")
    https://github.com/opengisch/QField/blob/master/src/core/qgismobileapp.cpp
    """
    # Something provided here just to get rid of warnings
    qmlRegisterType(QgsProjectStub, "org.qgis", 1, 0, "Foo")
    # qmlRegisterUncreatableType
    # qmlRegisterSingletonType


def pytest_addoption(parser: "Parser") -> None:
    parser.addini(
        _QFIELD_IMPORT_DIR_KEY,
        "Absolute path to a directory containing QField imports. "
        "Usually src/qml/imports in QField source code.",
        default=os.getenv(_QFIELD_IMPORTS_DIR_ENV),
    )


def _get_qfied_import_path(request: "SubRequest") -> Path:
    qfield_import_path_config = request.config.getini(
        _QFIELD_IMPORT_DIR_KEY
    ) or os.getenv(_QFIELD_IMPORTS_DIR_ENV)
    if not qfield_import_path_config:
        raise ValueError(
            f"No {_QFIELD_IMPORT_DIR_KEY} ini value "
            f"or {_QFIELD_IMPORTS_DIR_ENV} environment variable set!"
        )

    qfield_import_path = Path(qfield_import_path_config)
    if not qfield_import_path.exists():
        raise ValueError(f"{qfield_import_path} does not exist!")

    if not (qfield_import_path / "Theme").exists():
        raise ValueError(f"{qfield_import_path / 'Theme'} does not exist!")
    return qfield_import_path


def _embed_qml_window_in_qgis_main_window(
    qgis_parent: QWidget,
    qgis_canvas: "QgsMapCanvas",
    qml_overlay_widget: QQuickWidget,
) -> None:
    # Keep QML as a toolbar strip and show QGIS canvas below it.
    # Overlay transparency is not reliable across all Qt backends.
    central_widget = QWidget(parent=qgis_parent)
    central_layout = QVBoxLayout(central_widget)
    central_layout.setContentsMargins(0, 0, 0, 0)
    central_layout.setSpacing(0)

    qml_overlay_widget.setParent(central_widget)
    qml_overlay_widget.setFocusPolicy(Qt.FocusPolicy.NoFocus)
    qml_overlay_widget.setSizePolicy(
        QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed
    )
    qml_overlay_widget.setFixedHeight(56)
    central_layout.addWidget(qml_overlay_widget, 0)

    qgis_canvas.setParent(central_widget)
    qgis_canvas.setSizePolicy(
        QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding
    )
    central_layout.addWidget(qgis_canvas, 1)

    qgis_parent.setCentralWidget(central_widget)


def _load_qml_overlay_widget(
    qml_engine: QQmlApplicationEngine,
    main_window_qml_path: Path,
    parent: QWidget,
) -> tuple[QQuickWidget, QObject]:
    qml_overlay_widget = QQuickWidget(qml_engine, parent)
    surface_format = qml_overlay_widget.format()
    surface_format.setAlphaBufferSize(8)
    qml_overlay_widget.setFormat(surface_format)
    qml_overlay_widget.setResizeMode(QQuickWidget.ResizeMode.SizeRootObjectToView)
    qml_overlay_widget.setAutoFillBackground(False)
    qml_overlay_widget.setClearColor(Qt.GlobalColor.transparent)
    qml_overlay_widget.setAttribute(Qt.WidgetAttribute.WA_TranslucentBackground, True)
    qml_overlay_widget.setStyleSheet("background: transparent;")
    qml_overlay_widget.setSource(QUrl.fromLocalFile(str(main_window_qml_path)))

    if qml_overlay_widget.status() == QQuickWidget.Status.Error:
        errors = [error.toString() for error in qml_overlay_widget.errors()]
        raise ValueError(f"QML file {main_window_qml_path} failed to load: {errors}")

    root_object = qml_overlay_widget.rootObject()
    if root_object is None:
        raise ValueError(f"QML file {main_window_qml_path} did not load successfully!")
    return qml_overlay_widget, root_object



# Copyright (C) 2026 pytest-qfield Contributors.
#
#
# This file is part of pytest-qfield.
#
# pytest-qfield is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# pytest-qfield is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pytest-qfield.  If not, see <https://www.gnu.org/licenses/>.
import shutil
from contextlib import suppress
from pathlib import Path
from typing import TYPE_CHECKING, Any, Protocol, cast

import pytest_qgis.utils
from PyQt6.QtCore import QObject, QPointF, Qt, QtMsgType, QUrl
from PyQt6.QtQml import QQmlComponent
from PyQt6.QtQuick import QQuickItem
from qgis.core import QgsProject

if TYPE_CHECKING:
    from PyQt6.QtQml import QQmlApplicationEngine
    from pytestqt.logging import _QtMessageCapture
    from pytestqt.qtbot import QtBot

    from pytest_qfield.stub_interface.qfield_stubs import QFieldAppInterfaceStub

QML_JS_QOBJECT_TEMPLATE = """
import QtQml
import "{js_file_name}" as Logic

QtObject {{
    function call({params}) {{
        return Logic.{function_name}({params})
    }}
}}

"""


class JsQObject(Protocol):
    """
    Protocol for JavaScript QObject with a call method.
    """

    @staticmethod
    def call(*args: Any) -> Any:  # noqa: ANN401
        """
        Function to call the JavaScript function with the given parameters.
        :param args: Parameters to call the function with.
        :return: A possible return value from the function
        """
        ...


class QFieldBot:
    """
    Helper class for testing QField QML files.
    """

    def __init__(
        self,
        qfield_iface: "QFieldAppInterfaceStub",
        qml_engine: "QQmlApplicationEngine",
        qtbot: "QtBot",
        qtlog: "_QtMessageCapture",
        tmp_path: "Path",
    ) -> None:
        self.iface = qfield_iface
        self._qml_engine = qml_engine
        self.qtbot = qtbot
        self.qtlog = qtlog
        self._tmp_path = tmp_path
        self._plugin_loaded = False
        self._plugin_root_object: QObject | None = None

    def show_window(self) -> None:
        """
        Show the QField application window.
        """
        self.iface.show()
        QgsProject.instance().layersAdded.connect(
            self.sync_canvas_layers_to_project_layer_tree
        )

    def open_project(self, project_file: Path) -> None:
        """
        Open a QField project file.

        If you load the plugin before opening the project, remember to load it with
        emit_load_project_ended=False.

        :param project_file: Path to the QGIS project file to open.
        """
        QgsProject.instance().clear()
        with suppress(TypeError):
            QgsProject.instance().layersAdded.disconnect(
                self.sync_canvas_layers_to_project_layer_tree
            )
        try:
            with self.qtbot.waitSignal(QgsProject.instance().readProject):
                assert QgsProject.instance().read(str(project_file))
            self.sync_canvas_layers_to_project_layer_tree()
            self.emit_load_project_ended()
        finally:
            QgsProject.instance().layersAdded.connect(
                self.sync_canvas_layers_to_project_layer_tree
            )

    def load_plugin(
        self,
        qfield_plugin_qml_file: Path,
        raise_if_warnings: bool = True,
        emit_load_project_ended: bool = True,
    ) -> None:
        """
        Load a QField plugin QML file into the QML engine and wait for it to be loaded.

        :param qfield_plugin_qml_file: Path to the QML file to load.
        :param raise_if_warnings: Whether to raise an exception
                if any warnings or errors are logged during loading.
        :param emit_load_project_ended: Whether to emit the loadProjectEnded
                signal on iface.
        """

        self._plugin_root_object = self.load_qml(
            qfield_plugin_qml_file, raise_if_warnings
        )
        self.iface.set_qml_root(self._plugin_root_object)
        self._plugin_loaded = True
        if emit_load_project_ended:
            self.emit_load_project_ended()

    def emit_load_project_ended(self) -> None:
        """Emit the loadProjectEnded signal on iface after initial setup."""
        self.iface.loadProjectEnded.emit()

    def load_js_function(
        self,
        js_file: Path,
        name: str,
        params: list[str],
        extra_files: list[Path] | None = None,
    ) -> JsQObject:
        """
        Load js function from file as a QObject with function "call".
        Call object.call to execute the function.

        :param js_file: Path to the JavaScript file to load.
        :param name: Name of the function to load.
        :param params: List of parameter names for the function.
        :param extra_files: List of extra files to load.
        :return: JsQObject with the loaded function
        """

        for file in [js_file, *(extra_files if extra_files else [])]:
            shutil.copy(file, self._tmp_path / file.name)
        qml_code = QML_JS_QOBJECT_TEMPLATE.format(
            js_file_name=js_file.name, function_name=name, params=", ".join(params)
        )
        main_qml = self._tmp_path / "main.qml"
        main_qml.touch()
        main_qml.write_text(qml_code)

        component = QQmlComponent(self._qml_engine)
        component.loadUrl(QUrl.fromLocalFile(str(main_qml)))
        js_object = component.create()
        if js_object is None:
            raise ValueError(
                "Could not create QObject with js function: ",
                [error.toString() for error in component.errors()],
            )
        return cast("JsQObject", js_object)

    def get_item(self, object_name: str) -> "QQuickItem":
        self._ensure_plugin_loaded()
        child = self._get_plugin_root_object().findChild(QQuickItem, object_name)
        if not child:
            raise RuntimeError(f"QML object {object_name} not found!")
        return child

    def click_item(
        self,
        item: "QQuickItem",
        mouse_button: Qt.MouseButton = Qt.MouseButton.LeftButton,
    ) -> None:
        center = QPointF(item.width(), item.height()) / 2
        self.qtbot.mouseClick(
            item.window(), mouse_button, pos=item.mapToScene(center).toPoint()
        )

    def load_qml(self, qml_file: Path, raise_if_warnings: bool = True) -> QObject:
        """
        Load a QML file into the QML engine and wait for it to be loaded.

        :param qml_file: Path to the QML file to load.
        :param raise_if_warnings: Whether to raise an exception
                if any warnings or errors are logged during loading.
        """
        initial_log_count = len(self.qtlog.records)
        initial_root_count = len(self._qml_engine.rootObjects())

        if not qml_file.exists():
            raise FileNotFoundError(f"QML file {qml_file} does not exist!")
        with self.qtbot.waitSignal(self._qml_engine.objectCreated):
            self._qml_engine.load(str(qml_file))

        new_logs = [
            (m.type, m.message.strip())
            for m in self.qtlog.records[initial_log_count:]
            if m.type
            in (QtMsgType.QtWarningMsg, QtMsgType.QtCriticalMsg, QtMsgType.QtFatalMsg)
        ]
        if raise_if_warnings and new_logs:
            raise ValueError(f"QML file {qml_file} failed to load:\n{new_logs}")

        if not len(self._qml_engine.rootObjects()) > initial_root_count:
            raise ValueError(f"QML file {qml_file} did not load successfully!")

        return self._qml_engine.rootObjects()[-1]

    def set_map_crs_based_on_layers(self) -> None:
        """Set map crs based on layers of the project."""
        # Copied here for convenience
        pytest_qgis.utils.set_map_crs_based_on_layers()

    def _ensure_plugin_loaded(self) -> None:
        if not self._plugin_loaded:
            raise ValueError("Plugin not loaded yet!")

    def _get_plugin_root_object(self) -> "QObject":
        if self._plugin_root_object is None:
            raise ValueError("Plugin not loaded yet!")
        return self._plugin_root_object

    def sync_canvas_layers_to_project_layer_tree(self, *_: Any) -> None:  # noqa: ANN401
        """
        Set map canvas layers in project layer tree order.

        This is alternative way compared to settings up the QgsLayerTreeMapCanvasBridge,
        which does not seem to cope well with web map services.
        """
        root = QgsProject.instance().layerTreeRoot()
        ordered_layers = []
        for layer in root.layerOrder():
            layer_node = root.findLayer(layer.id())
            if layer_node is None:
                continue
            is_visible = (
                layer_node.itemVisibilityChecked()
                if hasattr(layer_node, "itemVisibilityChecked")
                else layer_node.isVisible()
            )
            if is_visible:
                ordered_layers.append(layer)
        self.iface.qgis_map_canvas.setLayers(ordered_layers)
        self.iface.qgis_map_canvas.refreshAllLayers()



# Copyright (C) 2026 pytest-qfield Contributors.
#
#
# This file is part of pytest-qfield.
#
# pytest-qfield is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# pytest-qfield is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pytest-qfield.  If not, see <https://www.gnu.org/licenses/>.

# Resource object code (Python 3)
# Created by: object code
# Created by: The Resource Compiler for Qt version 5.15.18
# WARNING! All changes made in this file will be lost!

from PyQt6 import QtCore

qt_resource_data = b"\
\x00\x00\x005\
\x0a\
import QtQuick\x0ai\
mport QtQuick.Co\
ntrols\x0a\x0aButton {\
\x0a}\x0a\x0a\
"

qt_resource_name = b"\
\x00\x03\
\x00\x00x<\
\x00q\
\x00m\x00l\
\x00\x08\
\x0c\xa7[|\
\x00t\
\x00e\x00s\x00t\x00.\x00q\x00m\x00l\
"

qt_resource_struct = b"\
\x00\x00\x00\x00\x00\x02\x00\x00\x00\x01\x00\x00\x00\x01\
\x00\x00\x00\x00\x00\x00\x00\x00\
\x00\x00\x00\x00\x00\x02\x00\x00\x00\x01\x00\x00\x00\x02\
\x00\x00\x00\x00\x00\x00\x00\x00\
\x00\x00\x00\x0c\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\
\x00\x00\x01\x9c\xcf\x0dB\x88\
"


def qInitResources():
    QtCore.qRegisterResourceData(
        0x03, qt_resource_struct, qt_resource_name, qt_resource_data
    )


def qCleanupResources():
    QtCore.qUnregisterResourceData(
        0x03, qt_resource_struct, qt_resource_name, qt_resource_data
    )


qInitResources()
